#ifndef _SEGWAYRMP400_H
#define _SEGWAYRMP400_H

#include "segway_rmp200.h"
#include "ftdiserver.h"
#include "threadserver.h"
#include "eventserver.h"
#include "mutex.h"
#include <vector>
#include <string>
#include <iostream>

const unsigned int NUM_SEGWAY_200 = 2;

typedef enum {front_segway=0,rear_segway} segway_id_t;

struct TSegwayRMP400Status
{
  TSegwayRMP200Status rmp200[NUM_SEGWAY_200];
};

class CSegwayRMP400
{
  private:
    std::vector<CSegwayRMP200> segways;
    /**
     * \brief forward displacement
     *
     * This value has the displacement of the whole platform in meters
     * TODO: should be a thread periodically updating this value
     */
    double forward_displacement;

    /**
     * \brief yaw displacement
     *
     * This value has the rotation of the whole platform in revolutions.
     * TODO: should be a thread periodically updating this value
     */
    double yaw_displacement;

    /**
     * \brief yaw rate
     *
     * This value has the yaw rate of the platform in degrees per second.
     * TODO: should be a thread periodically updating this value
     */
    double yaw_rate;
    /**
     * \brief
     *
     */
    bool connected;
    /**
     * \brief
     *
     */
    std::string front_serial;
    /**
     * \brief
     *
     */
    std::string rear_serial;
    /**
     * \brief
     *
     */
    CThreadServer *thread_server;
    /**
     * \brief
     *
     */
    std::string monitor_thread_id;
    /**
     * \brief
     *
     */
    CEventServer *event_server;
    /**
     * \brief
     *
     */
    std::string finish_thread_event_id;
    /**
     * \brief
     *
     */
    std::string front_cable_disconnect_event_id;
    /**
     * \brief
     *
     */
    std::string rear_cable_disconnect_event_id;
    /**
     * \brief
     *
     */
    std::string front_power_off_event_id;
    /**
     * \brief
     *
     */
    std::string rear_power_off_event_id;
    /**
     * \brief
     *
     */
    std::string front_no_heartbeat_event_id;
    /**
     * \brief
     *
     */
    std::string rear_no_heartbeat_event_id;
    /**
     * \brief
     *
     */
    std::string new_status_event_id;
    /**
     * \brief
     *
     */
    CMutex status_access;
    /**
     * \brief
     *
     */
    TSegwayRMP400Status status;
    /* \brief aux function to check if platform is really doing the reset in
     *        integrators.
     */
    bool is_valid_reset_integrators();
  protected:
    /**
     * \brief
     *
     */
    static void *monitor_thread(void *param);
    /**
     * \brief
     *
     */
    void init_events_threads(void);
  public:
    /**
     * \brief
     *
     */
    CSegwayRMP400();
    /**
     * \brief
     *
     */
    CSegwayRMP400(const std::string &front, const std::string &rear);
    /**
     * \brief
     *
     */
    void connect(const std::string &front,const std::string &rear);

    /**
     * \brief function to return whole platform status
     *
     * It will return the two segway internal status
     */
    TSegwayRMP400Status get_status(void);

    /**
     * \brief function to return the total forward displacement
     *
     * This function returns the current forward displacement in meters. This
     * function only returns the value of the internal attribute, but it does
     * not access the hardware platform. This value is periodically updated by
     * the feedback thread.
     *
     * Both fordward displacement from segways are sum and a mean is calculated
     *
     * \return the current forward displacement in meters.
     *
     */
    double get_forward_displacement(void);

    /**
     * \brief function to return the total yaw displacement
     *
     * This function returns the current yaw displacement in revolutions per
     * second. This function only returns the value of the internal attribute,
     * but it does not access the hardware platform. This value is periodically
     * updated by the feedback thread.
     *
     * Both yaw displacement from segways are sum and a mean is calculated
     *
     * \return the current yaw displacement in revolutions per second.
     *
     */
    double get_yaw_displacement(void);

    /**
     * \brief function to return the yaw rate
     *
     * This function returns the current yaw rate in revolutions per second.
     * This function only returns the value of the internal attribute, but it
     * does not access the hardware platform. This value is periodically updated
     * by the feedback thread.
     *
     * Both yaw rate from segways are sum and a mean is calculated
     *
     * \return the current yaw rate in revolutions per second.
     *
     */
    double get_yaw_rate(void);

    /**
     * \brief function to reset integrators
     *
     * This function sets all encoder integrators to zero in both platforms.
     * Currently this function seems NOT to be reliable since hardware not
     * always is doing the proper work.
     */
    void reset_integrators();

    /**
     * \brief
     *
     */
    std::string get_front_cable_disconnected_event_id(void);
    /**
     * \brief
     *
     */
    std::string get_rear_cable_disconnected_event_id(void);
    /**
     * \brief
     *
     */
    std::string get_front_power_off_event_id(void);
    /**
     * \brief
     *
     */
    std::string get_rear_power_off_event_id(void);
    /**
     * \brief
     *
     */
    std::string get_front_no_heartbeat_event_id(void);
    /**
     * \brief
     *
     */
    std::string get_rear_no_heartbeat_event_id(void);
    /**
     * \brief
     *
     */
    std::string get_new_status_event_id(void);

    /**
     * \brief function to move the platform
     *
     * This function sets the rotational and translational velocities to a desired value
     * vT is in m/sec, vR is in rev/sec
     *
     * Rotation about the center of the platform with skid steer is disabled.
     * Skid steering is not recommended since it produces significant tire wear and unnecessary battery consumption.
     * Odometry readings during skid steering are useless since they depend heavily on soil characteristics.
     * A minimum Instantaneous Center of Rotation (ICR) located on the side of the platform is enforced (0.54cm).
     */
    void move(double vT, double vR);

    /* \brief function to stop the platform
     *
     * This function stops the vehicle
     */
    void stop();

    /**
     * \brief
     *
     */
    void close();

    /**
     * \brief
     *
     */
    friend std::ostream& operator<< (std::ostream& out, CSegwayRMP400& segway);
    /**
     * \brief
     *
     */
    ~CSegwayRMP400();
};

#endif


